from typing import Optional, Sequence, Tuple, Union

import equinox as eqx
import equinox.nn as nn
import jax
import jax.nn as jnn
import jax.numpy as jnp
import jax.random as jrandom
from equinox.custom_types import Array

from ...layers import DropPath, MlpProjection, PatchEmbed
from ...utils import load_torch_weights, MODEL_URLS


class _VitAttention(eqx.Module):
    num_heads: int
    scale: float
    qkv: nn.Linear
    attn_drop: nn.Dropout
    proj: nn.Linear
    proj_drop: nn.Dropout

    def __init__(
        self,
        dim: int,
        num_heads: int = 8,
        qkv_bias: bool = False,
        qk_scale=None,
        attn_drop: float = 0.0,
        proj_drop: float = 0.0,
        *,
        key: "jax.random.PRNGKey" = None,
    ):
        """**Arguments:**

        - `dim`: The feature dimensions of the input.
        - `num_heads`: The number of equal parts to split the input along the `dim`.
        - `qkv_bias`: To add `bias` within the `qkv` computation.
        - `qk_scale`: For scaling the `query` `value` computation.
        - `attn_drop`: Dropout rate for the attention.
        - `proj_drop`: Dropout rate for the projection.
        - key: A `jax.random.PRNGKey` used to provide randomness for parameter
        initialisation. (Keyword only argument.)
        """
        super().__init__()
        keys = jrandom.split(key, 2)
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = qk_scale or head_dim**-0.5

        self.qkv = nn.Linear(dim, dim * 3, use_bias=qkv_bias, key=keys[0])
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim, key=keys[1])
        self.proj_drop = nn.Dropout(proj_drop)

    def __call__(self, x: Array, *, key: "jax.random.PRNGKey") -> Sequence[Array]:
        """**Arguments:**

        - `x`: The input `JAX` array.
        - `key`: Utilised by few layers in the network such as `Dropout` or `BatchNorm`.
        """
        N, C = x.shape
        keys = jrandom.split(key, 2)
        qkv = jax.vmap(self.qkv)(x)
        qkv = jnp.reshape(qkv, (N, 3, self.num_heads, C // self.num_heads))
        qkv = jnp.transpose(qkv, axes=(1, 2, 0, 3))
        q, k, v = jnp.split(qkv, indices_or_sections=3)

        attn = (q @ jnp.transpose(k, (0, 1, 3, 2))) * self.scale
        attn = jnn.softmax(attn, axis=-1)
        attn = self.attn_drop(attn, key=keys[0])

        x = jnp.reshape(jnp.transpose((attn @ v), axes=(0, 2, 1, 3)), (N, C))
        x = jax.vmap(self.proj)(x)
        x = self.proj_drop(x, key=keys[1])
        return x, attn


class _VitBlock(eqx.Module):
    norm1: eqx.Module
    attn: _VitAttention
    drop_path: DropPath
    norm2: eqx.Module
    mlp: MlpProjection

    def __init__(
        self,
        dim,
        num_heads,
        mlp_ratio=4.0,
        qkv_bias=False,
        qk_scale=None,
        drop=0.0,
        attn_drop=0.0,
        drop_path=0.0,
        act_layer=jnn.gelu,
        norm_layer=nn.LayerNorm,
        *,
        key: "jax.random.PRNGKey",
    ):
        """**Arguments**

        - `dim`: The feature dimensions of the input.
        - `num_heads`: The number of equal parts to split the input along the `dim`.
        - `mlp_ratio`: For computing hidden dimension of the `mlp` (=`dim * mlp_ratio`).
        - `qkv_bias`: To add `bias` within the `qkv` computation.
        - `qk_scale`: For scaling the `query` `value` computation.
        - `drop`: Dropout rate for the `mlp`.
        - `attn_drop`: Dropout rate for the attention.
        - `proj_drop`: Dropout rate for the projection.
        - `act_layer`: Activation applied on to the intermediate outputs.
        - `norm_layer`: Normalisation applied to the intermediate outputs.
        - key: A `jax.random.PRNGKey` used to provide randomness for parameter
        initialisation. (Keyword only argument.)
        """
        super().__init__()
        keys = jrandom.split(key, 2)
        self.norm1 = norm_layer(dim)
        self.attn = _VitAttention(
            dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            qk_scale=qk_scale,
            attn_drop=attn_drop,
            proj_drop=drop,
            key=keys[0],
        )
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = MlpProjection(
            in_features=dim,
            hidden_features=mlp_hidden_dim,
            act_layer=act_layer,
            drop=drop,
            key=keys[1],
        )

    def __call__(
        self, x: Array, return_attention=False, *, key: "jax.random.PRNGKey"
    ) -> Array:
        """**Arguments:**

        - `x`: The input. Should be a JAX array with `3` channels.
        - `return_attention`: For returning the self-attention computed by the block.
        - `key`: Utilised by few layers in the network such as `Dropout` or `BatchNorm`.
        """
        keys = jrandom.split(key, 4)
        y = jax.vmap(self.norm1)(x)
        y, attn = self.attn(y, key=keys[0])
        if return_attention:
            return attn
        x = x + self.drop_path(y, key=keys[1])
        y = jax.vmap(self.norm2)(x)
        y = jax.vmap(self.mlp)(y, key=jrandom.split(keys[2], x.shape[0]))
        x = x + self.drop_path(y, key=keys[3])
        return x


class VisionTransformer(eqx.Module):
    """Vision Transformer ported from https://github.com/facebookresearch/dino/blob/main/vision_transformer.py"""

    num_features: int
    cls_token: jnp.ndarray
    pos_embed: jnp.ndarray
    patch_embed: PatchEmbed
    pos_drop: nn.Dropout
    blocks: Sequence[_VitBlock]
    norm: eqx.Module
    fc: nn.Linear
    inference: bool

    def __init__(
        self,
        img_size: Union[int, Tuple[int]] = 224,
        patch_size: Union[int, Tuple[int]] = 16,
        in_chans: int = 3,
        num_classes: int = 0,
        embed_dim: int = 768,
        depth: int = 12,
        num_heads: int = 12,
        mlp_ratio: float = 4.0,
        qkv_bias: bool = True,
        qk_scale=None,
        drop_rate=0.0,
        attn_drop_rate=0.0,
        drop_path_rate=0.0,
        norm_layer=nn.LayerNorm,
        *,
        key: Optional["jax.random.PRNGKey"] = None,
    ):
        """**Arguments:**

        - `img_size`: The size of the input image. Defaults to `(224, 224)`
        - `patch_size`: Size of the patch to construct from the input image. Defaults to `(16, 16)`
        - `in_chans`: Number of input channels. Defaults to `3`
        - `num_classes`: Number of classes in the classification task.
                         Also controls the final output shape `(num_classes,)`. If `num_classes=0`,
                         then the final layer is replaced with `nn.Identity`
        - `embed_dim`: The dimension of the resulting embedding of the patch. Defaults to `768`
        - `depth`: Number of `VitBlocks` in the network
        - `num_heads`: The number of equal parts to split the input along the `dim`
        - `mlp_ratio`: For computing hidden dimension of the `mlp`
        - `qkv_bias`: To add `bias` within the `qkv` computation
        - `qk_scale`: For scaling the `query` `value` computation
        - `drop_rate`: Dropout rate used within the `MlpProjection`
        - `attn_drop_rate`: Dropout rate used within the attention modules
        - `drop_path_rate`: Dropout rate used within `VitBlock`s
        - `norm_layer`: Normalisation applied to the intermediate outputs. Defaults to `equinox.nn.LayerNorm`
        - `key`: A `jax.random.PRNGKey` used to provide randomness for parameter
        initialisation. (Keyword only argument.)

        """

        super().__init__()
        if key is None:
            key = jrandom.PRNGKey(0)
        keys = jrandom.split(key, depth + 3)
        self.inference = False
        self.num_features = embed_dim
        self.patch_embed = PatchEmbed(
            img_size=img_size,
            patch_size=patch_size,
            in_chans=in_chans,
            embed_dim=embed_dim,
        )
        num_patches = self.patch_embed.num_patches

        self.cls_token = jrandom.truncated_normal(
            keys[0], lower=-2, upper=2, shape=(1, embed_dim)
        )
        self.pos_embed = jrandom.truncated_normal(
            keys[1], lower=-2, upper=2, shape=(num_patches + 1, embed_dim)
        )
        self.pos_drop = nn.Dropout(p=drop_rate)
        dpr = jnp.linspace(0, drop_path_rate, depth)
        self.blocks = [
            _VitBlock(
                dim=embed_dim,
                num_heads=num_heads,
                mlp_ratio=mlp_ratio,
                qkv_bias=qkv_bias,
                qk_scale=qk_scale,
                drop=drop_rate,
                attn_drop=attn_drop_rate,
                drop_path=dpr[i],
                norm_layer=norm_layer,
                key=keys[i + 1],
            )
            for i in range(depth)
        ]
        self.norm = norm_layer(embed_dim)
        # Classifier head
        self.fc = (
            nn.Identity()
            if num_classes == 0
            else nn.Linear(embed_dim, num_classes, key=keys[-1])
        )
        # ToDo: Initialization scheme of the weights

    def __call__(self, x: Array, *, key: "jax.random.PRNGKey") -> Array:
        """**Arguments:**

        - `x`: The input `JAX` array
        - `key`: Required parameter. Utilised by few layers such as `Dropout` or `DropPath`
        """
        keys = jrandom.split(key, len(self.blocks))
        x = self.patch_embed(x)
        x = jnp.concatenate([self.cls_token, x], axis=0) + self.pos_embed
        for key_, blk in zip(keys, self.blocks):
            x = blk(x, key=key_)
        x = jax.vmap(self.norm)(x)
        return self.fc(x[0])

    def get_last_self_attention(self, x: Array, *, key: "jax.random.PRNGKey") -> Array:
        """**Arguments:**

        - `x`: The input `JAX` array
        - `key`: Utilised by few layers in the network such as `Dropout` or `BatchNorm`
        """
        if not self.inference:
            raise ValueError(
                "Model being evaluated outside inference mode. Try in inference mode."
            )
        keys = jrandom.split(key, len(self.blocks))
        x = self.patch_embed(x)
        x = jnp.concatenate([self.cls_token, x], axis=0) + self.pos_embed
        for i, blk in enumerate(self.blocks):
            if i < len(self.blocks) - 1:
                x = blk(x, key=keys[i])
            else:
                return blk(x, return_attention=True, key=key)


def vit_tiny(
    patch_size=16,
    embed_dim=192,
    depth=12,
    num_heads=3,
    mlp_ratio=4,
    pretrained=False,
    *,
    key: Optional["jax.random.PRNGKey"] = None,
    **kwargs,
):
    """**Arguments:**

    - `patch_size`: Size of the patch to construct from the input image
    - `embed_dim`: The dimension of the resulting embedding of the patch
    - `depth`: Number of `VitBlocks` in the network
    - `num_heads`: The number of equal parts to split the input along the `dim`
    - `mlp_ratio`: For computing hidden dimension of the `mlp`
    - `key`: A `jax.random.PRNGKey` used to provide randomness for parameter
        initialisation. (Keyword only argument.)
    - `kwargs`: Parameters passed on to the `VisionTransformer`
    """
    if pretrained:
        raise NotImplementedError("vit_tiny does not support pretrained mode.")
    model = VisionTransformer(
        patch_size=patch_size,
        embed_dim=embed_dim,
        depth=depth,
        num_heads=num_heads,
        mlp_ratio=mlp_ratio,
        key=key,
        **kwargs,
    )
    return model


def vit_small(
    patch_size=16,
    embed_dim=384,
    depth=12,
    num_heads=6,
    mlp_ratio=4,
    pretrained=False,
    *,
    key: Optional["jax.random.PRNGKey"] = None,
    **kwargs,
):
    """**Arguments:**

    - `patch_size`: Size of the patch to construct from the input image
    - `embed_dim`: The dimension of the resulting embedding of the patch
    - `depth`: Number of `VitBlocks` in the network
    - `num_heads`: The number of equal parts to split the input along the `dim`
    - `mlp_ratio`: For computing hidden dimension of the `mlp`
    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint
    - `key`: A `jax.random.PRNGKey` used to provide randomness for parameter
        initialisation. (Keyword only argument.)
    - `kwargs`: Parameters passed on to the `VisionTransformer`
    """
    model = VisionTransformer(
        patch_size=patch_size,
        embed_dim=embed_dim,
        depth=depth,
        num_heads=num_heads,
        mlp_ratio=mlp_ratio,
        key=key,
        **kwargs,
    )
    if pretrained:
        if not isinstance(model.fc, nn.Identity):
            raise ValueError(
                "Currently ViT only support DINO weights which require"
                " no classification layer. Try again with num_classes=0."
            )

        model_name = f"vit_small_patch{patch_size}_224_dino"
        model = load_torch_weights(model, url=MODEL_URLS[model_name])
    return model


def vit_base(
    patch_size=16,
    embed_dim=768,
    depth=12,
    num_heads=12,
    mlp_ratio=4,
    pretrained=False,
    *,
    key: Optional["jax.random.PRNGKey"] = None,
    **kwargs,
):
    """**Arguments:**

    - `patch_size`: Size of the patch to construct from the input image
    - `embed_dim`: The dimension of the resulting embedding of the patch
    - `depth`: Number of `VitBlocks` in the network
    - `num_heads`: The number of equal parts to split the input along the `dim`
    - `mlp_ratio`: For computing hidden dimension of the `mlp`
    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint
    - `key`: A `jax.random.PRNGKey` used to provide randomness for parameter
        initialisation. (Keyword only argument.)
    - `kwargs`: Parameters passed on to the `VisionTransformer`
    """
    model = VisionTransformer(
        patch_size=patch_size,
        embed_dim=embed_dim,
        depth=depth,
        num_heads=num_heads,
        mlp_ratio=mlp_ratio,
        key=key,
        **kwargs,
    )
    if pretrained:
        if not isinstance(model.fc, nn.Identity):
            raise ValueError(
                "Currently ViT only support DINO weights which require"
                " no classification layer. Try again with num_classes=0."
            )

        model_name = f"vit_base_patch{patch_size}_224_dino"
        model = load_torch_weights(model, url=MODEL_URLS[model_name])
    return model



from typing import Any, Callable, List, Optional, Sequence, Type, Union

import equinox as eqx
import equinox.experimental as eqex
import equinox.nn as nn
import jax
import jax.nn as jnn
import jax.numpy as jnp
import jax.random as jrandom
from equinox.custom_types import Array

from ...utils import load_torch_weights, MODEL_URLS


def _conv3x3(in_planes, out_planes, stride=1, groups=1, dilation=1, key=None):
    """3x3 convolution with padding"""
    return nn.Conv2d(
        in_planes,
        out_planes,
        kernel_size=3,
        stride=stride,
        padding=dilation,
        groups=groups,
        use_bias=False,
        dilation=dilation,
        key=key,
    )


def _conv1x1(in_planes, out_planes, stride=1, key=None):
    """1x1 convolution"""
    return nn.Conv2d(
        in_planes, out_planes, kernel_size=1, stride=stride, use_bias=False, key=key
    )


class _ResNetBasicBlock(eqx.Module):
    expansion: int
    conv1: eqx.Module
    bn1: eqx.Module
    relu: Callable
    conv2: eqx.Module
    bn2: eqx.Module
    downsample: eqx.Module
    stride: int

    def __init__(
        self,
        inplanes,
        planes,
        stride=1,
        downsample=None,
        groups=1,
        base_width=64,
        dilation=1,
        norm_layer=None,
        key=None,
    ):
        super(_ResNetBasicBlock, self).__init__()
        if norm_layer is None:
            norm_layer = eqex.BatchNorm
        if groups != 1 or base_width != 64:
            raise ValueError("BasicBlock only supports groups=1 and base_width=64")
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        keys = jrandom.split(key, 2)
        self.expansion = 1
        self.conv1 = _conv3x3(inplanes, planes, stride, key=keys[0])
        self.bn1 = norm_layer(planes, axis_name="batch")
        self.relu = jnn.relu
        self.conv2 = _conv3x3(planes, planes, key=keys[1])
        self.bn2 = norm_layer(planes, axis_name="batch")
        if downsample:
            self.downsample = downsample
        else:
            self.downsample = nn.Identity()
        self.stride = stride

    def __call__(
        self, x: Array, *, key: Optional["jax.random.PRNGKey"] = None
    ) -> Array:
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)
        out = self.conv2(out)
        out = self.bn2(out)
        identity = self.downsample(x)
        out += identity
        out = self.relu(out)

        return out


class _ResNetBottleneck(eqx.Module):
    # Bottleneck in torchvision places the stride for downsampling at 3x3 convolution(self.conv2)
    # while original implementation places the stride at the first 1x1 convolution(self.conv1)
    # according to "Deep residual learning for image recognition"https://arxiv.org/abs/1512.03385.
    # This variant is also known as ResNet V1.5 and improves accuracy according to
    # https://ngc.nvidia.com/catalog/model-scripts/nvidia:resnet_50_v1_5_for_pytorch.
    expansion: int
    conv1: eqx.Module
    bn1: eqx.Module
    conv2: eqx.Module
    bn2: eqx.Module
    conv3: eqx.Module
    bn3: eqx.Module
    relu: Callable
    downsample: eqx.Module
    stride: int

    def __init__(
        self,
        inplanes,
        planes,
        stride=1,
        downsample=None,
        groups=1,
        base_width=64,
        dilation=1,
        norm_layer=None,
        key=None,
    ):
        super(_ResNetBottleneck, self).__init__()
        if norm_layer is None:
            norm_layer = eqex.BatchNorm
        self.expansion = 4
        keys = jrandom.split(key, 3)
        width = int(planes * (base_width / 64.0)) * groups
        # Both self.conv2 and self.downsample layers downsample the input when stride != 1
        self.conv1 = _conv1x1(inplanes, width, key=keys[0])
        self.bn1 = norm_layer(width, axis_name="batch")
        self.conv2 = _conv3x3(width, width, stride, groups, dilation, key=keys[1])
        self.bn2 = norm_layer(width, axis_name="batch")
        self.conv3 = _conv1x1(width, planes * self.expansion, key=keys[2])
        self.bn3 = norm_layer(planes * self.expansion, axis_name="batch")
        self.relu = jnn.relu
        if downsample:
            self.downsample = downsample
        else:
            self.downsample = nn.Identity()
        self.stride = stride

    def __call__(self, x: Array, *, key: Optional["jax.random.PRNGKey"] = None):

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


EXPANSIONS = {_ResNetBasicBlock: 1, _ResNetBottleneck: 4}


class ResNet(eqx.Module):
    """A simple port of `torchvision.models.resnet`"""

    inplanes: int
    dilation: int
    groups: Sequence[int]
    base_width: int
    conv1: eqx.Module
    bn1: eqx.Module
    relu: jnn.relu
    maxpool: eqx.Module
    layer1: eqx.Module
    layer2: eqx.Module
    layer3: eqx.Module
    layer4: eqx.Module
    avgpool: eqx.Module
    fc: eqx.Module

    def __init__(
        self,
        block: Type[Union["_ResNetBasicBlock", "_ResNetBottleneck"]],
        layers: List[int],
        num_classes: int = 1000,
        groups: int = 1,
        width_per_group: int = 64,
        replace_stride_with_dilation: List[bool] = None,
        norm_layer: Any = None,
        *,
        key: Optional["jax.random.PRNGKey"] = None,
    ):
        """**Arguments:**

        - `block`: `Bottleneck` or `BasicBlock` for constructing the network
        - `layers`: A list containing number of `blocks` at different levels
        - `num_classes`: Number of classes in the classification task.
                        Also controls the final output shape `(num_classes,)`. Defaults to `1000`
        - `groups`: Number of groups to form along the feature depth. Defaults to `1`
        - `width_per_group`: Increases width of `block` by a factor of `width_per_group/64`.
        Defaults to `64`
        - `replace_stride_with_dilation`: Replacing `2x2` strides with dilated convolution. Defaults to None
        - `norm_layer`: Normalisation to be applied on the inputs. Defaults to `BatchNorm`
        - `key`: A `jax.random.PRNGKey` used to provide randomness for parameter
        initialisation. (Keyword only argument.)

        ??? Failure "Exceptions:"

            - `NotImplementedError`: If a `norm_layer` other than `equinox.experimental.BatchNorm` is used
            - `ValueError`: If `replace_stride_with_convolution` is not `None` or a `3-tuple`

        """
        super(ResNet, self).__init__()
        if not norm_layer:
            norm_layer = eqex.BatchNorm

        if eqex.BatchNorm != norm_layer:
            raise NotImplementedError(
                f"{type(norm_layer)} is not currently supported. Use `eqx.experimental.BatchNorm` instead."
            )
        if key is None:
            key = jrandom.PRNGKey(0)

        keys = jrandom.split(key, 6)
        self.inplanes = 64
        self.dilation = 1
        if replace_stride_with_dilation is None:
            # each element in the tuple indicates if we should replace
            # the 2x2 stride with a dilated convolution instead
            replace_stride_with_dilation = [False, False, False]
        if len(replace_stride_with_dilation) != 3:
            raise ValueError(
                "replace_stride_with_dilation should be None "
                "or a 3-element tuple, got {}".format(replace_stride_with_dilation)
            )
        self.groups = groups
        self.base_width = width_per_group
        self.conv1 = nn.Conv2d(
            3,
            self.inplanes,
            kernel_size=7,
            stride=2,
            padding=3,
            use_bias=False,
            key=keys[0],
        )
        self.bn1 = norm_layer(input_size=self.inplanes, axis_name="batch")
        self.relu = jnn.relu
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0], norm_layer, key=keys[1])
        self.layer2 = self._make_layer(
            block,
            128,
            layers[1],
            norm_layer,
            stride=2,
            dilate=replace_stride_with_dilation[0],
            key=keys[2],
        )
        self.layer3 = self._make_layer(
            block,
            256,
            layers[2],
            norm_layer,
            stride=2,
            dilate=replace_stride_with_dilation[1],
            key=keys[3],
        )
        self.layer4 = self._make_layer(
            block,
            512,
            layers[3],
            norm_layer,
            stride=2,
            dilate=replace_stride_with_dilation[2],
            key=keys[4],
        )
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(512 * EXPANSIONS[block], num_classes, key=keys[5])

    def _make_layer(
        self, block, planes, blocks, norm_layer, stride=1, dilate=False, key=None
    ):
        keys = jrandom.split(key, blocks + 1)
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * EXPANSIONS[block]:
            downsample = nn.Sequential(
                [
                    _conv1x1(
                        self.inplanes, planes * EXPANSIONS[block], stride, key=keys[0]
                    ),
                    norm_layer(planes * EXPANSIONS[block], axis_name="batch"),
                ]
            )

        layers = []
        layers.append(
            block(
                self.inplanes,
                planes,
                stride,
                downsample,
                self.groups,
                self.base_width,
                previous_dilation,
                norm_layer,
                key=keys[1],
            )
        )
        self.inplanes = planes * EXPANSIONS[block]
        for block_idx in range(1, blocks):
            layers.append(
                block(
                    self.inplanes,
                    planes,
                    groups=self.groups,
                    base_width=self.base_width,
                    dilation=self.dilation,
                    norm_layer=norm_layer,
                    key=keys[block_idx + 1],
                )
            )

        return nn.Sequential(layers)

    def __call__(self, x: Array, *, key: "jax.random.PRNGKey") -> Array:
        """**Arguments:**

        - `x`: The input. Should be a JAX array with `3` channels
        - `key`: Required parameter. Utilised by few layers such as `Dropout` or `DropPath`
        """
        if key is None:
            raise RuntimeError("The model requires a PRNGKey.")
        keys = jrandom.split(key, 6)
        x = self.conv1(x, key=keys[0])
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x, key=keys[1])
        x = self.layer2(x, key=keys[2])
        x = self.layer3(x, key=keys[3])
        x = self.layer4(x, key=keys[4])

        x = self.avgpool(x)
        x = jnp.ravel(x)
        x = self.fc(x, key=keys[5])

        return x


def _resnet(block, layers, **kwargs):
    model = ResNet(block, layers, **kwargs)
    return model


def resnet18(pretrained=False, **kwargs) -> ResNet:
    r"""ResNet-18 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.
    """
    model = _resnet(_ResNetBasicBlock, [2, 2, 2, 2], **kwargs)
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS["resnet18"])
    return model


def resnet34(pretrained=False, **kwargs) -> ResNet:
    r"""ResNet-34 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.

    """
    model = _resnet(_ResNetBasicBlock, [3, 4, 6, 3], **kwargs)
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS["resnet34"])
    return model


def resnet50(pretrained=False, **kwargs) -> ResNet:
    r"""ResNet-50 model from
    [Deep Residual Learning for Image Recognition](https://arxiv.org/pdf/1512.03385.pdf)

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.

    """
    model = _resnet(_ResNetBottleneck, [3, 4, 6, 3], **kwargs)
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS["resnet50"])
    return model


def resnet101(pretrained=False, **kwargs) -> ResNet:
    r"""ResNet-101 model from
    [Deep Residual Learning for Image Recognition](https://arxiv.org/pdf/1512.03385.pdf)

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.

    """
    model = _resnet(_ResNetBottleneck, [3, 4, 23, 3], **kwargs)
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS["resnet101"])
    return model


def resnet152(pretrained=False, **kwargs) -> ResNet:
    r"""ResNet-152 model from
    [Deep Residual Learning for Image Recognition](https://arxiv.org/pdf/1512.03385.pdf)

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.

    """
    model = _resnet(_ResNetBottleneck, [3, 8, 36, 3], **kwargs)
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS["resnet152"])
    return model


def resnext50_32x4d(pretrained=False, **kwargs) -> ResNet:
    r"""ResNeXt-50 32x4d model from
    [Aggregated Residual Transformation for Deep Neural Networks](https://arxiv.org/pdf/1611.05431.pdf)

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.

    """
    kwargs["groups"] = 32
    kwargs["width_per_group"] = 4
    model = _resnet(_ResNetBottleneck, [3, 4, 6, 3], **kwargs)
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS["resnext50_32x4d"])
    return model


def resnext101_32x8d(pretrained=False, **kwargs) -> ResNet:
    r"""ResNeXt-101 32x8d model from
    [Aggregated Residual Transformation for Deep Neural Networks](https://arxiv.org/pdf/1611.05431.pdf)

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.

    """
    kwargs["groups"] = 32
    kwargs["width_per_group"] = 8
    model = _resnet(_ResNetBottleneck, [3, 4, 23, 3], **kwargs)
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS["resnext101_32x8d"])
    return model


def wide_resnet50_2(pretrained=False, **kwargs) -> ResNet:
    r"""Wide ResNet-50-2 model from
    [Wide Residual Networks](https://arxiv.org/pdf/1605.07146.pdf)
    The model is the same as ResNet except for the bottleneck number of channels
    which is twice larger in every block. The number of channels in outer 1x1
    convolutions is the same, e.g. last block in ResNet-50 has 2048-512-2048
    channels, and in Wide ResNet-50-2 has 2048-1024-2048.

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.

    """
    kwargs["width_per_group"] = 64 * 2
    model = _resnet(_ResNetBottleneck, [3, 4, 6, 3], **kwargs)
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS["wide_resnet50_2"])
    return model


def wide_resnet101_2(pretrained=False, **kwargs) -> ResNet:
    r"""Wide ResNet-101-2 model from
    [Wide Residual Networks](https://arxiv.org/pdf/1605.07146.pdf)
    The model is the same as ResNet except for the bottleneck number of channels
    which is twice larger in every block. The number of channels in outer 1x1
    convolutions is the same, e.g. last block in ResNet-50 has 2048-512-2048
    channels, and in Wide ResNet-50-2 has 2048-1024-2048.

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint.

    """
    kwargs["width_per_group"] = 64 * 2
    model = _resnet(_ResNetBottleneck, [3, 4, 23, 3], **kwargs)
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS["wide_resnet101_2"])
    return model



from functools import partial
from typing import Any, Callable, List, Optional, Sequence

import equinox as eqx
import equinox.nn as nn
import jax
import jax.nn as jnn
import jax.numpy as jnp
import jax.random as jrandom
from equinox.custom_types import Array

from ...layers import ConvNormActivation, DropPath, LayerNorm2d, Linear2d
from ...utils import load_torch_weights, MODEL_URLS


class CNBlock(eqx.Module):
    layer_scale: jnp.ndarray
    block: nn.Sequential
    stochastic_depth: DropPath

    def __init__(
        self,
        dim,
        layer_scale: float,
        stochastic_depth_prob: float,
        norm_layer: Optional[Callable[..., eqx.Module]] = LayerNorm2d,
        *,
        key=None,
    ) -> None:
        super().__init__()
        if norm_layer is None:
            norm_layer = partial(nn.LayerNorm, eps=1e-6)

        keys = jrandom.split(key, 4)
        self.block = nn.Sequential(
            [
                nn.Conv2d(
                    dim,
                    dim,
                    kernel_size=7,
                    padding=3,
                    groups=dim,
                    use_bias=True,
                    key=keys[0],
                ),
                norm_layer(dim),
                Linear2d(
                    in_features=dim, out_features=4 * dim, use_bias=True, key=keys[1]
                ),
                nn.Lambda(jnn.gelu),
                Linear2d(
                    in_features=4 * dim, out_features=dim, use_bias=True, key=keys[2]
                ),
            ]
        )
        self.layer_scale = jnp.asarray(
            jnp.ones(shape=(dim, 1, 1)) * layer_scale, dtype=jnp.float32
        )
        self.stochastic_depth = DropPath(p=stochastic_depth_prob, mode="local")

    def __call__(self, x: Array, *, key: "jax.random.PRNGKey") -> Array:
        """**Arguments:**

        - `x`: The input `JAX` array
        - `key`: Utilised by few layers in the network such as `Dropout` or `DropPath`
        """
        keys = jrandom.split(key, 2)
        result = self.layer_scale * self.block(x, key=keys[0])
        result = self.stochastic_depth(result, key=keys[1])
        result += x
        return result


class _CNBlockConfig:
    # Stores information listed at Section 3 of the ConvNeXt paper
    def __init__(
        self,
        input_channels: int,
        out_channels: Optional[int],
        num_layers: int,
    ) -> None:
        self.input_channels = input_channels
        self.out_channels = out_channels
        self.num_layers = num_layers

    def __repr__(self) -> str:
        s = self.__class__.__name__ + "("
        s += "input_channels={input_channels}"
        s += ", out_channels={out_channels}"
        s += ", num_layers={num_layers}"
        s += ")"
        return s.format(**self.__dict__)


class ConvNeXt(eqx.Module):
    """A simple port of `torchvision.models.convnext`."""

    features: nn.Sequential
    avgpool: nn.AdaptiveAvgPool2d
    classifier: nn.Sequential

    def __init__(
        self,
        block_setting: Sequence["_CNBlockConfig"],
        stochastic_depth_prob: float = 0.0,
        layer_scale: float = 1e-6,
        num_classes: int = 1000,
        block: Optional["eqx.Module"] = None,
        norm_layer: Optional["eqx.Module"] = None,
        *,
        key: "jax.random.PRNGKey" = None,
    ) -> None:
        """

        - `block_setting`: Configuration of the computational blocks
        - `stochastic_depth_prob`: Probability of dropping a sample along channels
        - `layer_scale`: Scale applied to the output of computational stem
        - `num_classes`: Number of classes in the classification task.
                         Also controls the final output shape `(num_classes,)`
        - `block`: The block type used within the network
        - `norm_layer`: Normalisation applied to the intermediate outputs
        - `key`: A `jax.random.PRNGKey` used to provide randomness for parameter
        initialisation. (Keyword only argument.)
        """
        super().__init__()

        if not block_setting:
            raise ValueError("The block_setting should not be empty")
        elif not (
            isinstance(block_setting, Sequence)
            and all([isinstance(s, _CNBlockConfig) for s in block_setting])
        ):
            raise TypeError("The block_setting should be List[CNBlockConfig]")

        if key is None:
            key = jrandom.PRNGKey(0)
        keys = jrandom.split(key, 2)

        if block is None:
            block = CNBlock

        if norm_layer is None:
            norm_layer = partial(LayerNorm2d, eps=1e-6)

        layers: List[eqx.Module] = []

        # Stem
        firstconv_output_channels = block_setting[0].input_channels
        layers.append(
            ConvNormActivation(
                in_channels=3,
                out_channels=firstconv_output_channels,
                kernel_size=4,
                stride=4,
                padding=0,
                norm_layer=norm_layer,
                activation_layer=None,
                use_bias=True,
                key=keys[0],
            )
        )

        total_stage_blocks = sum(cnf.num_layers for cnf in block_setting)
        stage_block_id = 0
        for cnf in block_setting:
            # Bottlenecks
            stage: List[eqx.Module] = []
            for _ in range(cnf.num_layers):
                keys = jrandom.split(keys[1], 2)
                # adjust stochastic depth probability based on the depth of the stage block
                sd_prob = (
                    stochastic_depth_prob * stage_block_id / (total_stage_blocks - 1.0)
                )
                stage.append(
                    block(cnf.input_channels, layer_scale, sd_prob, key=keys[0])
                )
                stage_block_id += 1
            layers.append(nn.Sequential(stage))
            if cnf.out_channels is not None:
                keys = jrandom.split(keys[1], 2)
                # Downsampling
                layers.append(
                    nn.Sequential(
                        [
                            norm_layer(cnf.input_channels),
                            nn.Conv2d(
                                cnf.input_channels,
                                cnf.out_channels,
                                kernel_size=2,
                                stride=2,
                                key=keys[0],
                            ),
                        ]
                    )
                )

        self.features = nn.Sequential(layers)
        self.avgpool = nn.AdaptiveAvgPool2d(1)

        lastblock = block_setting[-1]
        lastconv_output_channels = (
            lastblock.out_channels
            if lastblock.out_channels is not None
            else lastblock.input_channels
        )
        self.classifier = nn.Sequential(
            [
                norm_layer(lastconv_output_channels),
                nn.Lambda(jnp.ravel),
                nn.Linear(lastconv_output_channels, num_classes, key=keys[1]),
            ]
        )

    def __call__(self, x: Array, *, key: "jax.random.PRNGKey") -> Array:
        """**Arguments:**

        - `x`: The input `JAX` array.
        - `key`: Required parameter. Utilised by few layers such as `Dropout` or `DropPath`.
        """
        x = self.features(x, key=key)
        x = self.avgpool(x)
        x = self.classifier(x)
        return x


def _convnext(
    arch: str,
    block_setting: List[_CNBlockConfig],
    stochastic_depth_prob: float,
    pretrained: bool,
    **kwargs: Any,
) -> ConvNeXt:
    model = ConvNeXt(
        block_setting, stochastic_depth_prob=stochastic_depth_prob, **kwargs
    )
    if pretrained:
        if arch not in MODEL_URLS:
            raise ValueError(f"No checkpoint is available for model type {arch}")
        model = load_torch_weights(model, url=MODEL_URLS[arch])
    return model


def convnext_tiny(*, pretrained: bool = False, **kwargs: Any) -> ConvNeXt:
    r"""ConvNeXt Tiny model architecture from the
    `"A ConvNet for the 2020s" <https://arxiv.org/abs/2201.03545>`_ paper.

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint
    """
    block_setting = [
        _CNBlockConfig(96, 192, 3),
        _CNBlockConfig(192, 384, 3),
        _CNBlockConfig(384, 768, 9),
        _CNBlockConfig(768, None, 3),
    ]
    stochastic_depth_prob = kwargs.pop("stochastic_depth_prob", 0.1)
    return _convnext(
        "convnext_tiny",
        block_setting=block_setting,
        stochastic_depth_prob=stochastic_depth_prob,
        pretrained=pretrained,
        **kwargs,
    )


def convnext_small(*, pretrained: bool = False, **kwargs: Any) -> ConvNeXt:
    r"""ConvNeXt Small model architecture from the
    [A ConvNet for the 2020s](https://arxiv.org/abs/2201.03545) paper.

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint
    """
    block_setting = [
        _CNBlockConfig(96, 192, 3),
        _CNBlockConfig(192, 384, 3),
        _CNBlockConfig(384, 768, 27),
        _CNBlockConfig(768, None, 3),
    ]
    stochastic_depth_prob = kwargs.pop("stochastic_depth_prob", 0.4)
    return _convnext(
        "convnext_small", block_setting, stochastic_depth_prob, pretrained, **kwargs
    )


def convnext_base(*, pretrained: bool = False, **kwargs: Any) -> ConvNeXt:
    r"""ConvNeXt Base model architecture from the
    [A ConvNet for the 2020s](https://arxiv.org/abs/2201.03545) paper.

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint
    """
    block_setting = [
        _CNBlockConfig(128, 256, 3),
        _CNBlockConfig(256, 512, 3),
        _CNBlockConfig(512, 1024, 27),
        _CNBlockConfig(1024, None, 3),
    ]
    stochastic_depth_prob = kwargs.pop("stochastic_depth_prob", 0.5)
    return _convnext(
        "convnext_base", block_setting, stochastic_depth_prob, pretrained, **kwargs
    )


def convnext_large(*, pretrained: bool = False, **kwargs: Any) -> ConvNeXt:
    r"""ConvNeXt Large model architecture from the
    [A ConvNet for the 2020s](https://arxiv.org/abs/2201.03545) paper.

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint
    """
    block_setting = [
        _CNBlockConfig(192, 384, 3),
        _CNBlockConfig(384, 768, 3),
        _CNBlockConfig(768, 1536, 27),
        _CNBlockConfig(1536, None, 3),
    ]
    stochastic_depth_prob = kwargs.pop("stochastic_depth_prob", 0.5)
    return _convnext(
        "convnext_large", block_setting, stochastic_depth_prob, pretrained, **kwargs
    )



import equinox as eqx
import jax
import jax.numpy as jnp
import jax.random as jrandom
from equinox.custom_types import Array


class DropPath(eqx.Module):
    """Effectively dropping a sample from the call.
    Often used inside a network along side a residual connection.
    Equivalent to `torchvision.stochastic_depth`."""

    p: float
    inference: bool
    mode: str

    def __init__(self, p: float = 0.0, inference: bool = False, mode="global"):
        """**Arguments:**

        - `p`: The probability to drop a sample entirely during forward pass
        - `inference`: Defaults to `False`. If `True`, then the input is returned unchanged
        This may be toggled with `equinox.tree_inference`
        - `mode`: Can be set to `global` or `local`. If `global`, the whole input is dropped or retained.
                If `local`, then the decision on each input unit is computed independently. Defaults to `global`

        !!! note

            For `mode = local`, an input `(channels, dim_0, dim_1, ...)` is reshaped and transposed to
            `(channels, dims).transpose()`. For each `dim x channels` element,
            the decision to drop/keep is made independently.

        """
        self.p = p
        self.inference = inference
        self.mode = mode

    def __call__(self, x, *, key: "jax.random.PRNGKey") -> Array:
        """**Arguments:**

        - `x`: An any-dimensional JAX array to drop
        - `key`: A `jax.random.PRNGKey` used to provide randomness for calculating
            which elements to dropout. (Keyword only argument.)
        """
        if self.inference or self.p == 0.0:
            return x
        if key is None:
            raise RuntimeError(
                "DropPath requires a key when running in non-deterministic mode. Did you mean to enable inference?"
            )

        keep_prob = 1 - self.p
        if self.mode == "global":
            return x * jrandom.bernoulli(key, p=keep_prob)
        else:
            return x * jnp.expand_dims(
                jrandom.bernoulli(key, p=keep_prob, shape=[x.shape[0]]).reshape(-1),
                axis=[i for i in range(1, len(x.shape))],
            )



import warnings
from functools import partial
from typing import Any, Callable, List, Optional

import equinox as eqx
import equinox.nn as nn
import jax
import jax.nn as jnn
import jax.numpy as jnp
import jax.random as jr
from equinox.custom_types import Array

from ...layers import DropPath, LayerNorm2d, Linear2d, MlpProjection
from ...utils import load_torch_weights, MODEL_URLS


def _func_dropout(x, p, key):
    q = 1 - p
    mask = jr.bernoulli(key, q, x.shape)
    return jnp.where(mask, x / q, 0)


def _patch_merging_pad(x: Array) -> Array:
    _, H, W = x.shape
    x = jnp.pad(x, ((0, 0), (0, H % 2), (0, W % 2)))
    x0 = x[:, 0::2, 0::2]  # ... H/2 W/2 C
    x1 = x[:, 1::2, 0::2]  # ... H/2 W/2 C
    x2 = x[:, 0::2, 1::2]  # ... H/2 W/2 C
    x3 = x[:, 1::2, 1::2]  # ... H/2 W/2 C
    x = jnp.concatenate([x0, x1, x2, x3], axis=0)  # 4*C H/2 W/2
    return x


def _get_relative_position_bias(
    relative_position_bias_table: Array,
    relative_position_index: Array,
    window_size: List[int],
) -> Array:
    N = window_size[0] * window_size[1]
    relative_position_bias = relative_position_bias_table[relative_position_index]  # type: ignore[index]
    relative_position_bias = jnp.reshape(relative_position_bias, (N, N, -1))
    relative_position_bias = jnp.transpose(relative_position_bias, (2, 0, 1))
    return relative_position_bias


class _PatchMerging(eqx.Module):
    reduction: Linear2d
    norm: Callable

    def __init__(
        self,
        dim: int,
        norm_layer: Callable = LayerNorm2d,
        *,
        key: "jax.random.PRNGKey" = None,
    ):
        super().__init__()
        self.norm = norm_layer(4 * dim)
        self.reduction = Linear2d(4 * dim, 2 * dim, use_bias=False, key=key)

    def __call__(self, x: Array, *, key: "jax.random.PRNGKey") -> Array:
        x = _patch_merging_pad(x)
        x = self.norm(x)
        x = self.reduction(x)  # ... H/2 W/2 2*C
        return x


class _PatchMergingV2(eqx.Module):
    reduction: Linear2d
    norm: Callable

    def __init__(
        self,
        dim: int,
        norm_layer: Callable = LayerNorm2d,
        *,
        key: "jax.random.PRNGKey" = None,
    ):
        super().__init__()
        self.norm = norm_layer(2 * dim)
        self.reduction = Linear2d(4 * dim, 2 * dim, use_bias=False, key=key)

    def __call__(self, x: Array, *, key: "jax.random.PRNGKey") -> Array:
        x = _patch_merging_pad(x)
        x = self.reduction(x)  # ... H/2 W/2 2*C
        x = self.norm(x)
        return x


def _shifted_window_attention(
    x: Array,
    qkv_weight: Array,
    proj_weight: Array,
    relative_position_bias: Array,
    window_size: List[int],
    num_heads: int,
    shift_size: List[int],
    attention_dropout: float = 0.0,
    dropout: float = 0.0,
    qkv_bias: Optional[Array] = None,
    proj_bias: Optional[Array] = None,
    logit_scale: Optional["jax.numpy.ndarray"] = None,
    key: "jax.random.PRNGKey" = None,
):
    """Like `torchvision.models.swin_transformer.shifted_window_attention`."""
    x = jnp.transpose(x, (1, 2, 0))
    H, W, C = x.shape
    # TODO: Support flexible padding
    # pad feature maps to multiples of window size
    # pad_r = jnp.mod((window_size[1] - jnp.mod(W, window_size[1])), window_size[1])
    # pad_b = jnp.mod((window_size[0] - jnp.mod(H, window_size[0])), window_size[0])
    # x = jnp.pad(x, ((0, pad_r), (0, pad_b), (0, 0)))
    pad_H, pad_W, _ = x.shape

    shift_size = shift_size.copy()
    # If window size is larger than feature size, there is no need to shift window
    if window_size[0] >= pad_H:
        shift_size[0] = 0
    if window_size[1] >= pad_W:
        shift_size[1] = 0

    # cyclic shift
    if sum(shift_size) > 0:
        x = jnp.roll(x, shift=(-shift_size[0], -shift_size[1]), axis=(0, 1))

    # partition windows
    num_windows = (pad_H // window_size[0]) * (pad_W // window_size[1])
    x = jnp.reshape(
        x,
        (
            pad_H // window_size[0],
            window_size[0],
            pad_W // window_size[1],
            window_size[1],
            C,
        ),
    )
    x = jnp.reshape(
        jnp.transpose(x, (0, 2, 1, 3, 4)),
        (num_windows, window_size[0] * window_size[1], C),
    )  # nW, Ws*Ws, C

    # multi-head attention
    if logit_scale is not None and qkv_bias is not None:
        qkv_bias = qkv_bias.copy()
        length = qkv_bias.size // 3
        set_zero_slice = lambda c, a: (c.at[a].set(0), a)
        qkv_bias, _ = jax.lax.scan(
            set_zero_slice, qkv_bias, jnp.arange(length, 2 * length)
        )  # qkv_bias[length:2*length].zero_()
    qkv = (
        jnp.matmul(x, jnp.transpose(qkv_weight)) + qkv_bias
    )  # F.linear(x, qkv_weight, qkv_bias)
    qkv = jnp.transpose(
        jnp.reshape(qkv, (x.shape[0], x.shape[1], 3, num_heads, C // num_heads)),
        (2, 0, 3, 1, 4),
    )
    q, k, v = qkv[0], qkv[1], qkv[2]
    if logit_scale is not None:
        # cosine attention
        attn = (q / jnp.linalg.norm(q, ord=2, axis=0)) @ jnp.transpose(
            (k / jnp.linalg.norm(k, ord=2, axis=0)), (0, 1, 3, 2)
        )
        # attn = F.normalize(q, dim=-1) @ F.normalize(k, dim=-1).transpose(-2, -1)
        logit_scale = jnp.exp(
            jax.lax.clamp(min=-jnp.inf, x=logit_scale, max=jnp.log(100.0))
        )
        attn = attn * logit_scale
    else:
        q = q * (C // num_heads) ** -0.5
        attn = q @ jnp.transpose(k, (0, 1, 3, 2))
    # add relative position bias
    attn = attn + relative_position_bias

    if sum(shift_size) > 0:
        # generate attention mask
        attn_00 = jnp.ones((pad_H - window_size[0], pad_W - window_size[1])) * 0
        attn_01 = jnp.ones((pad_H - window_size[0], window_size[1] - shift_size[1])) * 1
        attn_02 = jnp.ones((pad_H - window_size[0], shift_size[1])) * 2
        attn_10 = jnp.ones((window_size[0] - shift_size[0], pad_W - window_size[1])) * 3
        attn_11 = (
            jnp.ones((window_size[0] - shift_size[0], window_size[1] - shift_size[1]))
            * 4
        )
        attn_12 = jnp.ones((window_size[0] - shift_size[0], shift_size[1])) * 5
        attn_20 = jnp.ones((shift_size[0], pad_W - window_size[1])) * 6
        attn_21 = jnp.ones((shift_size[0], window_size[1] - shift_size[1])) * 7
        attn_22 = jnp.ones((shift_size[0], shift_size[1])) * 8

        attn_mask = jnp.concatenate(
            jnp.concatenate(
                (
                    jnp.concatenate((attn_00, attn_01, attn_02), axis=1),
                    jnp.concatenate((attn_10, attn_11, attn_12), axis=1),
                    jnp.concatenate((attn_20, attn_21, attn_22), axis=1),
                )
            ),
            axis=0,
        )

        attn_mask = jnp.reshape(
            attn_mask,
            (
                pad_H // window_size[0],
                window_size[0],
                pad_W // window_size[1],
                window_size[1],
            ),
        )
        attn_mask = jnp.reshape(
            jnp.transpose(attn_mask, (0, 2, 1, 3)),
            (num_windows, window_size[0] * window_size[1]),
        )
        attn_mask = jnp.expand_dims(attn_mask, axis=1) - jnp.expand_dims(
            attn_mask, axis=2
        )
        attn_mask = jnp.where(attn_mask == 0, 0.0, -100.0)
        attn = jnp.reshape(
            attn,
            (x.shape[0] // num_windows, num_windows, num_heads, x.shape[1], x.shape[1]),
        )
        attn = attn + jnp.expand_dims(attn_mask, axis=1)
        attn = jnp.reshape(attn, (-1, num_heads, x.shape[1], x.shape[1]))

    attn = jnn.softmax(attn, axis=-1)

    attn = _func_dropout(attn, p=attention_dropout, key=key)

    x = jnp.reshape(
        jnp.transpose((attn @ v), (0, 2, 1, 3)), (x.shape[0], x.shape[1], C)
    )
    x = x @ jnp.transpose(proj_weight, (1, 0)) + proj_bias
    x = _func_dropout(x, p=dropout, key=key)

    # reverse windows
    x = jnp.reshape(
        x,
        (
            pad_H // window_size[0],
            pad_W // window_size[1],
            window_size[0],
            window_size[1],
            C,
        ),
    )
    x = jnp.reshape(jnp.transpose(x, (0, 2, 1, 3, 4)), (pad_H, pad_W, C))

    # reverse cyclic shift
    if sum(shift_size) > 0:
        x = jnp.roll(x, shift=(shift_size[0], shift_size[1]), axis=(0, 1))

    x = jnp.reshape(
        jnp.take(jnp.reshape(x, (-1, C)), indices=jnp.arange(H * W), axis=0), (H, W, C)
    )
    return jnp.transpose(x, (2, 0, 1))


class _ShiftedWindowAttention(eqx.Module):

    window_size: List[int]
    shift_size: List[int]
    num_heads: int
    attention_dropout: float
    dropout: float
    relative_position_bias_table: jnp.ndarray
    relative_position_index: jnp.ndarray
    qkv: nn.Linear
    proj: nn.Linear

    def __init__(
        self,
        dim: int,
        window_size: List[int],
        shift_size: List[int],
        num_heads: int,
        qkv_bias: bool = True,
        proj_bias: bool = True,
        attention_dropout: float = 0.0,
        dropout: float = 0.0,
        *,
        key: "jax.random.PRNGKey" = None,
    ):
        super().__init__()
        if len(window_size) != 2 or len(shift_size) != 2:
            raise ValueError("window_size and shift_size must be of length 2")

        keys = jr.split(key, 3)

        self.window_size = window_size
        self.shift_size = shift_size
        self.num_heads = num_heads
        self.attention_dropout = attention_dropout
        self.dropout = dropout

        self.qkv = Linear2d(dim, dim * 3, use_bias=qkv_bias, key=keys[0])
        self.proj = Linear2d(dim, dim, use_bias=proj_bias, key=keys[1])

        self.relative_position_bias_table = self.define_relative_position_bias_table(
            key=keys[2]
        )
        self.relative_position_index = self.define_relative_position_index()

    def define_relative_position_bias_table(self, key):
        return jr.truncated_normal(
            key=key,
            shape=(
                (2 * self.window_size[0] - 1) * (2 * self.window_size[1] - 1),
                self.num_heads,
            ),
            lower=2,
            upper=2,
        )

    def define_relative_position_index(self):
        # get pair-wise relative position index for each token inside the window
        coords_h = jnp.arange(self.window_size[0])
        coords_w = jnp.arange(self.window_size[1])
        coords = jnp.stack(jnp.meshgrid(coords_h, coords_w, indexing="ij"))  # 2, Wh, Ww
        coords_flatten = jax.vmap(jnp.ravel)(coords)
        relative_coords = (
            coords_flatten[:, :, None] - coords_flatten[:, None, :]
        )  # 2, Wh*Ww, Wh*Ww
        relative_coords = jnp.transpose(relative_coords, (1, 2, 0))  # Wh*Ww, Wh*Ww, 2
        jnp.stack(
            (
                (relative_coords[:, :, 0] + self.window_size[0] - 1)
                * 2
                * self.window_size[1]
                - 1,
                relative_coords[:, :, 1] + self.window_size[1] - 1,
            ),
            axis=-1,
        )
        relative_position_index = jnp.ravel(relative_coords.sum(axis=-1))  # Wh*Ww*Wh*Ww
        return relative_position_index

    def get_relative_position_bias(self) -> Array:
        return _get_relative_position_bias(
            self.relative_position_bias_table, self.relative_position_index, self.window_size  # type: ignore[arg-type]
        )

    def __call__(self, x: Array, *, key: "jax.random.PRNGKey") -> Array:
        """**Arguments:**

        - `x`: The input array of shape `(H, W, C)`
        - `key`: Required parameter. Utilised by few layers such as `Dropout` or `DropPath`

        **Returns:**

        An array of shape `(H, W, C)`
        """
        relative_position_bias = self.get_relative_position_bias()
        return _shifted_window_attention(
            x,
            self.qkv.weight,
            self.proj.weight,
            relative_position_bias,
            self.window_size,
            self.num_heads,
            shift_size=self.shift_size,
            attention_dropout=self.attention_dropout,
            dropout=self.dropout,
            qkv_bias=self.qkv.bias,
            proj_bias=self.proj.bias,
            key=key,
        )


class _ShiftedWindowAttentionV2(eqx.Module):

    window_size: List[int]
    shift_size: List[int]
    num_heads: int
    attention_dropout: float
    dropout: float
    logit_scale: jnp.ndarray
    relative_position_bias_table: jnp.ndarray
    relative_position_index: jnp.ndarray
    qkv: nn.Linear
    proj: nn.Linear
    cpb_mlp: nn.Sequential

    def __init__(
        self,
        dim: int,
        window_size: List[int],
        shift_size: List[int],
        num_heads: int,
        qkv_bias: bool = True,
        proj_bias: bool = True,
        attention_dropout: float = 0.0,
        dropout: float = 0.0,
        *,
        key: "jax.random.PRNGKey" = None,
    ):
        super().__init__()
        if len(window_size) != 2 or len(shift_size) != 2:
            raise ValueError("window_size and shift_size must be of length 2")

        keys = jr.split(key, 3)

        self.window_size = window_size
        self.shift_size = shift_size
        self.num_heads = num_heads
        self.attention_dropout = attention_dropout
        self.dropout = dropout

        self.qkv = Linear2d(dim, dim * 3, use_bias=qkv_bias, key=keys[0])
        self.proj = Linear2d(dim, dim, use_bias=proj_bias, key=keys[1])

        self.relative_position_bias_table = self.define_relative_position_bias_table(
            key=keys[2]
        )
        self.relative_position_index = self.define_relative_position_index()

        self.logit_scale = jnp.log(10 * jnp.ones((num_heads, 1, 1)))
        # mlp to generate continuous relative position bias
        self.cpb_mlp = nn.Sequential(
            [
                nn.Lambda(partial(jnp.transpose, axes=(2, 0, 1))),
                Linear2d(2, 512, use_bias=True, key=keys[1]),
                nn.Lambda(jnn.relu),
                Linear2d(512, num_heads, use_bias=False, key=keys[2]),
                nn.Lambda(partial(jnp.transpose, axes=(0, 1, 2))),
            ]
        )
        if qkv_bias:
            length = self.qkv.bias.size // 3
            set_zero_slice = lambda c, a: (c.at[a].set(0), a)
            qkv_bias, _ = jax.lax.scan(
                set_zero_slice, self.qkv.bias, jnp.arange(length, 2 * length)
            )
            self.qkv = eqx.tree_at(lambda l: l.bias, self.qkv, qkv_bias)

    def define_relative_position_index(self):
        # get pair-wise relative position index for each token inside the window
        coords_h = jnp.arange(self.window_size[0])
        coords_w = jnp.arange(self.window_size[1])
        coords = jnp.stack(jnp.meshgrid(coords_h, coords_w, indexing="ij"))  # 2, Wh, Ww
        coords_flatten = jax.vmap(jnp.ravel)(coords)
        relative_coords = (
            coords_flatten[:, :, None] - coords_flatten[:, None, :]
        )  # 2, Wh*Ww, Wh*Ww
        relative_coords = jnp.transpose(relative_coords, (1, 2, 0))  # Wh*Ww, Wh*Ww, 2
        jnp.stack(
            (
                (relative_coords[:, :, 0] + self.window_size[0] - 1)
                * 2
                * self.window_size[1]
                - 1,
                relative_coords[:, :, 1] + self.window_size[1] - 1,
            ),
            axis=-1,
        )
        relative_position_index = jnp.ravel(relative_coords.sum(axis=-1))  # Wh*Ww*Wh*Ww
        return relative_position_index

    def define_relative_position_bias_table(self, key):
        # get relative_coords_table
        relative_coords_h = jnp.arange(
            -(self.window_size[0] - 1), self.window_size[0], dtype=jnp.float32
        )
        relative_coords_w = jnp.arange(
            -(self.window_size[1] - 1), self.window_size[1], dtype=jnp.float32
        )
        relative_coords_table = jnp.stack(
            jnp.meshgrid(relative_coords_h, relative_coords_w, indexing="ij")
        )
        relative_coords_table = jnp.transpose(relative_coords_table, (1, 2, 0))

        relative_coords_table = jnp.stack(
            [
                relative_coords_table[:, :, 0] / self.window_size[0] - 1,
                relative_coords_table[:, :, 1] / self.window_size[1] - 1,
            ],
            axis=-1,
        )
        relative_coords_table = 8 * relative_coords_table  # normalize to -8, 8
        relative_coords_table = (
            jnp.sign(relative_coords_table)
            * jnp.log2(jnp.abs(relative_coords_table) + 1.0)
            / 3.0
        )
        return relative_coords_table

    def get_relative_position_bias(self) -> Array:
        relative_position_bias = _get_relative_position_bias(
            jnp.reshape(
                self.cpb_mlp(self.relative_position_bias_table), (-1, self.num_heads)
            ),
            self.relative_position_index,  # type: ignore[arg-type]
            self.window_size,
        )
        relative_position_bias = 16 * jnn.sigmoid(relative_position_bias)
        return relative_position_bias

    def __call__(self, x: Array, *, key: "jax.random.PRNGKey") -> Array:
        """**Arguments:**

        - `x`: The input array of shape `(H, W, C)`
        - `key`: Required parameter. Utilised by few layers such as `Dropout` or `DropPath`

        **Returns:**

        An array of shape `(H, W, C)`
        """
        relative_position_bias = self.get_relative_position_bias()
        return _shifted_window_attention(
            x,
            self.qkv.weight,
            self.proj.weight,
            relative_position_bias,
            self.window_size,
            self.num_heads,
            shift_size=self.shift_size,
            attention_dropout=self.attention_dropout,
            dropout=self.dropout,
            qkv_bias=self.qkv.bias,
            proj_bias=self.proj.bias,
            logit_scale=self.logit_scale,
            key=key,
        )


class _SwinTransformerBlock(eqx.Module):
    norm1: Callable
    attn: eqx.Module
    stochastic_depth: DropPath
    norm2: Callable
    mlp: MlpProjection

    def __init__(
        self,
        dim: int,
        num_heads: int,
        window_size: List[int],
        shift_size: List[int],
        mlp_ratio: float = 4.0,
        dropout: float = 0.0,
        attention_dropout: float = 0.0,
        stochastic_depth_prob: float = 0.0,
        norm_layer: Callable[..., eqx.Module] = LayerNorm2d,
        attn_layer: Callable[..., eqx.Module] = _ShiftedWindowAttention,
        *,
        key: "jax.random.PRNGKey" = None,
    ):
        super().__init__()
        keys = jr.split(key, 2)

        self.norm1 = norm_layer(dim)
        self.attn = attn_layer(
            dim,
            window_size,
            shift_size,
            num_heads,
            attention_dropout=attention_dropout,
            dropout=dropout,
            key=keys[0],
        )
        self.stochastic_depth = DropPath(stochastic_depth_prob, mode="local")
        self.norm2 = norm_layer(dim)
        self.mlp = MlpProjection(
            dim,
            int(dim * mlp_ratio),
            dim,
            lin_layer=Linear2d,
            act_layer=jnn.gelu,
            drop=dropout,
            key=keys[1],
        )

    def __call__(self, x: Array, *, key: "jax.random.PRNGKey") -> Array:
        keys = jr.split(key, 4)
        x = x + self.stochastic_depth(
            self.attn(self.norm1(x), key=keys[0]), key=keys[1]
        )
        x = x + self.stochastic_depth(self.mlp(self.norm2(x), key=keys[2]), key=keys[3])
        return x


class _SwinTransformerBlockV2(eqx.Module):

    norm1: Callable
    attn: eqx.Module
    stochastic_depth: DropPath
    norm2: Callable
    mlp: MlpProjection

    def __init__(
        self,
        dim: int,
        num_heads: int,
        window_size: List[int],
        shift_size: List[int],
        mlp_ratio: float = 4.0,
        dropout: float = 0.0,
        attention_dropout: float = 0.0,
        stochastic_depth_prob: float = 0.0,
        norm_layer: Callable[..., eqx.Module] = LayerNorm2d,
        attn_layer: Callable[..., eqx.Module] = _ShiftedWindowAttentionV2,
        *,
        key: "jax.random.PRNGKey" = None,
    ):
        super().__init__()
        keys = jr.split(key, 2)

        self.norm1 = norm_layer(dim)
        self.attn = attn_layer(
            dim,
            window_size,
            shift_size,
            num_heads,
            attention_dropout=attention_dropout,
            dropout=dropout,
            key=keys[0],
        )
        self.stochastic_depth = DropPath(stochastic_depth_prob, mode="local")
        self.norm2 = norm_layer(dim)
        self.mlp = MlpProjection(
            dim,
            int(dim * mlp_ratio),
            dim,
            lin_layer=Linear2d,
            act_layer=jnn.gelu,
            drop=dropout,
            key=keys[1],
        )

    def __call__(self, x: Array, *, key: "jax.random.PRNGKey") -> Array:
        keys = jr.split(key, 4)
        x = x + self.stochastic_depth(
            self.norm1(self.attn(x, key=keys[0])), key=keys[1]
        )
        x = x + self.stochastic_depth(self.norm2(self.mlp(x, key=keys[2])), key=keys[3])
        return x


class SwinTransformer(eqx.Module):
    """A simple port of `torchvision.models.swin_transformer`."""

    features: nn.Sequential
    norm: Callable
    avgpool: nn.AdaptiveAvgPool2d
    head: nn.Linear

    def __init__(
        self,
        patch_size: List[int],
        embed_dim: int,
        depths: List[int],
        num_heads: List[int],
        window_size: List[int],
        mlp_ratio: float = 4.0,
        dropout: float = 0.0,
        attention_dropout: float = 0.0,
        stochastic_depth_prob: float = 0.1,
        num_classes: int = 1000,
        norm_layer: Callable = None,
        block: "eqx.Module" = None,
        downsample_layer: "eqx.Module" = None,
        *,
        key: Optional["jax.random.PRNGKey"] = None,
    ):
        """**Arguments:**

        - `patch_size`: Size of the patch to construct from the input image. Defaults to `(4, 4)`
        - `embed_dim`: The dimension of the resulting embedding of the patch
        - `depths`: Depth of each Swin Transformer layer
        - `num_heads`: Number of attention heads in different layers
        - `window_size`: Window size
        - `mlp_ratio`:  Ratio of mlp hidden dim to embedding dim. Defaults to `4.0`
        - `dropout`: Dropout rate. Defaults to `0.0`
        - `attention_dropout`: Attention dropout rate. Defaults to `0.0`
        - `stochastic_depth_prob`:  Stochastic depth rate. Defaults to `0.1`
        - `num_classes`:  Number of classes in the classification task.
                         Also controls the final output shape `(num_classes,)`
        - `norm_layer`: Normalisation applied to the intermediate outputs. Defaults to `LayerNorm2d`
        - `block`: The SwinTransformer-v1/v2 block to use. Defaults to `_SwinTransformerBlock` which is used in `v1`
        - `downsample_layer`: Downsample layer (patch merging). Defaults to `_PatchMerging` which is used in `v1`
        - `key`:  A `jax.random.PRNGKey` used to provide randomness for parameter
        initialisation. (Keyword only argument.)

        !!! warning "Note"

            Currently, input image dimensions are required to be divisible by the `window_size` for
            each level of `depth`. For example, input image shape `(3, 224, 224)` works for window_size `(7, 7)`
            and `(3, 256, 256)` works for `(8, 8)`.
        """
        super().__init__()
        if key is None:
            key = jr.PRNGKey(0)
        keys = jr.split(key, 2)

        if block is None:
            block = _SwinTransformerBlock
        if norm_layer is None:
            norm_layer = partial(LayerNorm2d, eps=1e-5)
        if downsample_layer is None:
            downsample_layer = _PatchMerging

        layers: List[eqx.Module] = []
        layers.append(
            nn.Sequential(
                [
                    nn.Conv2d(
                        3,
                        embed_dim,
                        kernel_size=(patch_size[0], patch_size[1]),
                        stride=(patch_size[0], patch_size[1]),
                        key=keys[0],
                    ),
                    norm_layer(embed_dim),
                ]
            )
        )

        total_stage_blocks = sum(depths)
        stage_block_id = 0
        # build SwinTransformer blocks
        for i_stage in range(len(depths)):
            stage: List[eqx.Module] = []
            dim = embed_dim * 2**i_stage
            for i_layer in range(depths[i_stage]):
                keys = jr.split(keys[1], 2)
                # adjust stochastic depth probability based on the depth of the stage block
                sd_prob = (
                    stochastic_depth_prob
                    * float(stage_block_id)
                    / (total_stage_blocks - 1)
                )
                stage.append(
                    block(
                        dim,
                        num_heads[i_stage],
                        window_size=window_size,
                        shift_size=[
                            0 if i_layer % 2 == 0 else w // 2 for w in window_size
                        ],
                        mlp_ratio=mlp_ratio,
                        dropout=dropout,
                        attention_dropout=attention_dropout,
                        stochastic_depth_prob=sd_prob,
                        norm_layer=norm_layer,
                        key=keys[0],
                    )
                )
                stage_block_id += 1
            layers.append(nn.Sequential(stage))
            # add patch merging layer
            if i_stage < (len(depths) - 1):
                keys = jr.split(keys[1], 2)
                layers.append(downsample_layer(dim, norm_layer, key=keys[0]))
        self.features = nn.Sequential(layers)

        num_features = embed_dim * 2 ** (len(depths) - 1)
        self.norm = norm_layer(num_features)
        self.avgpool = nn.AdaptiveAvgPool2d(1)
        self.head = nn.Linear(num_features, num_classes, key=keys[1])

    def __call__(self, x: Array, *, key: "jax.random.PRNGKey") -> Array:
        """**Arguments:**

        - `x`: The input `JAX` array
        - `key`: Required parameter. Utilised by few layers such as `Dropout` or `DropPath`
        """
        keys = jr.split(key, 2)
        x = self.features(x, key=keys[0])
        x = self.norm(x)
        x = self.avgpool(x)
        x = jnp.ravel(x)
        x = self.head(x, key=keys[1])
        return x


def _swin_transformer(
    arch: str,
    patch_size: List[int],
    embed_dim: int,
    depths: List[int],
    num_heads: List[int],
    window_size: List[int],
    stochastic_depth_prob: float,
    pretrained: bool,
    **kwargs: Any,
) -> SwinTransformer:

    warnings.warn(
        "Currently, dynamic padding of the input is not supported! "
        + "Please make sure that the input is a multiple of window_size."
    )

    model = SwinTransformer(
        patch_size=patch_size,
        embed_dim=embed_dim,
        depths=depths,
        num_heads=num_heads,
        window_size=window_size,
        stochastic_depth_prob=stochastic_depth_prob,
        **kwargs,
    )
    if pretrained:
        model = load_torch_weights(model, url=MODEL_URLS[arch])
    return model


def swin_t(pretrained: bool = False, **kwargs: Any) -> SwinTransformer:
    """
    Constructs a swin_tiny architecture from
    [Swin Transformer: Hierarchical Vision Transformer using Shifted Windows](https://arxiv.org/pdf/2103.14030).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint
    """

    return _swin_transformer(
        arch="swin_t",
        patch_size=[4, 4],
        embed_dim=96,
        depths=[2, 2, 6, 2],
        num_heads=[3, 6, 12, 24],
        window_size=[7, 7],
        stochastic_depth_prob=0.2,
        pretrained=pretrained,
        **kwargs,
    )


def swin_s(pretrained: bool = False, **kwargs: Any) -> SwinTransformer:
    """
    Constructs a swin_small architecture from
    [Swin Transformer: Hierarchical Vision Transformer using Shifted Windows](https://arxiv.org/pdf/2103.14030).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint
    """
    return _swin_transformer(
        arch="swin_s",
        patch_size=[4, 4],
        embed_dim=96,
        depths=[2, 2, 18, 2],
        num_heads=[3, 6, 12, 24],
        window_size=[7, 7],
        stochastic_depth_prob=0.3,
        pretrained=pretrained,
        **kwargs,
    )


def swin_b(pretrained: bool = False, **kwargs: Any) -> SwinTransformer:
    """
    Constructs a swin_base architecture from
    [Swin Transformer: Hierarchical Vision Transformer using Shifted Windows](https://arxiv.org/pdf/2103.14030).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint
    """

    return _swin_transformer(
        arch="swin_b",
        patch_size=[4, 4],
        embed_dim=128,
        depths=[2, 2, 18, 2],
        num_heads=[4, 8, 16, 32],
        window_size=[7, 7],
        stochastic_depth_prob=0.5,
        pretrained=pretrained,
        **kwargs,
    )


def swin_v2_t(pretrained: bool = False, **kwargs: Any) -> SwinTransformer:
    """
    Constructs a swin_v2_tiny architecture from
    [Swin Transformer V2: Scaling Up Capacity and Resolution](https://arxiv.org/pdf/2111.09883).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint
    """

    return _swin_transformer(
        arch="swin_v2_t",
        patch_size=[4, 4],
        embed_dim=96,
        depths=[2, 2, 6, 2],
        num_heads=[3, 6, 12, 24],
        window_size=[8, 8],
        stochastic_depth_prob=0.2,
        block=_SwinTransformerBlockV2,
        downsample_layer=_PatchMergingV2,
        pretrained=pretrained,
        **kwargs,
    )


def swin_v2_s(pretrained: bool = False, **kwargs: Any) -> SwinTransformer:
    """
    Constructs a swin_v2_small architecture from
    [Swin Transformer V2: Scaling Up Capacity and Resolution](https://arxiv.org/pdf/2111.09883).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint
    """

    return _swin_transformer(
        arch="swin_v2_s",
        patch_size=[4, 4],
        embed_dim=96,
        depths=[2, 2, 18, 2],
        num_heads=[3, 6, 12, 24],
        window_size=[8, 8],
        stochastic_depth_prob=0.3,
        block=_SwinTransformerBlockV2,
        pretrained=pretrained,
        downsample_layer=_PatchMergingV2,
        **kwargs,
    )


def swin_v2_b(pretrained: bool = False, **kwargs: Any) -> SwinTransformer:
    """
    Constructs a swin_v2_base architecture from
    [Swin Transformer V2: Scaling Up Capacity and Resolution](https://arxiv.org/pdf/2111.09883).

    **Arguments:**

    - `pretrained`: If `True`, the weights are loaded from `PyTorch` saved checkpoint
    """

    return _swin_transformer(
        arch="swin_v2_b",
        patch_size=[4, 4],
        embed_dim=128,
        depths=[2, 2, 18, 2],
        num_heads=[4, 8, 16, 32],
        window_size=[8, 8],
        stochastic_depth_prob=0.5,
        block=_SwinTransformerBlockV2,
        downsample_layer=_PatchMergingV2,
        pretrained=pretrained,
        **kwargs,
    )
